from pathlib import Path

from PIL import Image
import typer

from stega.image.text import encode, decode

app = typer.Typer()


def start_callback(value: int) -> int:
    if value < 0:
        raise typer.BadParameter("Must be greater than 0")
    return value


def every_px_callback(value: int) -> int:
    if value < 1:
        raise typer.BadParameter("Must be greater than 1")
    return value


def msg_callback(value: str) -> str:
    if not value:
        raise typer.BadParameter("Message must not be empty")
    return value


def key_callback(value: str) -> str:
    if not value:
        raise typer.BadParameter("Key must not be empty")
    return value


@app.callback(help='Stega-saurus text image steganography')
def text_image_callback():
    pass


@app.command(name='encode', help="Encode a message using an image output as a PNG")
def img_encode(
    in_image_path: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        help="Original image path to use in encoding"
    ),
    out_image_path: Path = typer.Argument(
        ...,
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Output image path to encode the message into (file extension ignored, image will be a png)"
    ),
    msg: str = typer.Argument(..., callback=msg_callback, help="Message to encode in to the image"),
):
    """
    Encode a message into a copy of an image.
    """
    typer.secho(f"stega-saurus text image steganography encode", fg=typer.colors.MAGENTA, bold=True)

    if in_image_path == out_image_path:
        typer.secho(
            f"In and out images paths are the same, this would overwrite the original file",
            fg=typer.colors.RED,
            bold=True,
            err=True
        )
        raise typer.Abort()

    typer.echo(typer.style("Encoding...", fg=typer.colors.CYAN, bold=True))

    with Image.open(in_image_path) as out_image:
        key = encode(out_image, msg)

        # Currently only supporting PNG for output file of encoding
        out_path_png = out_image_path.with_suffix('.png')
        out_image.save(out_path_png, format='png')

        typer.secho("Done encoding", fg=typer.colors.GREEN, bold=True)
        typer.secho(f"Decode key: {key}", fg=typer.colors.CYAN, bold=False)


@app.command(name='decode', help="Decode a message from an image")
def img_decode(
    in_image_path: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        help="Image path with encoded msg"),
    key: str = typer.Argument(..., callback=key_callback, help="Image path with encoded msg"),
):
    """
    Decode a message from an image.
    """
    typer.secho(f"stega-saurus text image steganography decode", fg=typer.colors.MAGENTA, bold=True)
    typer.secho("Decoding...", fg=typer.colors.CYAN, bold=True)

    with Image.open(in_image_path) as in_image:
        msg = decode(in_image, key)

        typer.secho("Done decoding", fg=typer.colors.GREEN, bold=True)
        typer.secho("Encoded message:", fg=typer.colors.CYAN, bold=True)
        typer.secho(msg)


if __name__ == '__main__':
    app()



from PIL import Image

RGB = (0, 1, 2)

DEFAULT_START = 0
DEFAULT_EVERY_PX = 1
STARTING_LEN_KEY_CHAR = 'a'


def _px_num(i: int, start: int, every_px: int, channel_index: int) -> int:
    return (start + i * every_px) + channel_index


def _xcoord(i: int, start: int, every_px: int, channel_index: int, image_width: int) -> int:
    return _px_num(i, start, every_px, channel_index) % (image_width)


def _ycoord(i: int, start: int, every_px: int, channel_index: int, image_width: int) -> int:
    return int(_px_num(i, start, every_px, channel_index) / (image_width))


def _coords(i: int, start: int, every_px: int, image_width: int) -> tuple:
    # return (x0, y0), (x1, y1), (x2, y2)
    return tuple([
        (_xcoord(i, start, every_px, channel_ind, image_width), _ycoord(i, start, every_px, channel_ind, image_width))
        for channel_ind in RGB
    ])


def _encode_digit_in_channel_val(digit: str, original_channel_val: int) -> int:
    channel_digits = [char for char in str(original_channel_val).zfill(3)]
    channel_digits[-1] = digit
    channel_val = int(''.join(channel_digits))

    if channel_val > 255:
        channel_val -= 10

    return channel_val


def _decode_digit_from_channel_val(channel_val: int) -> int:
    return channel_val % 10


def _msg_len_to_str(msg_len: int) -> str:
    len_str = str(msg_len)
    out_chars = [chr(int(digit_char) + ord(STARTING_LEN_KEY_CHAR)) for digit_char in len_str]
    return ''.join(out_chars)


def _str_to_msg_len(encoded_len: str) -> int:
    len_digit_chars = ''.join([str(ord(char) - ord(STARTING_LEN_KEY_CHAR)) for char in encoded_len])
    return int(''.join(len_digit_chars))


def encode(image: Image, msg: str, start: int = DEFAULT_START, every_px: int = DEFAULT_EVERY_PX) -> str:
    pxa = image.load()
    msg_bytes = msg.encode()
    byte_vals = [b for b in msg_bytes]

    # TODO: CHECK image dims vs msg length and needed encoding size

    for i, val in enumerate(byte_vals):
        coords = _coords(i, start, every_px, image.width)
        pixels = [pxa[x, y] for x, y in coords]
        digits = [d for d in str(val).zfill(3)]

        rgb_ind = i % len(RGB)

        for digit, pixel, pixel_coords in zip(digits, pixels, coords):
            new_channel_val = _encode_digit_in_channel_val(digit, pixel[rgb_ind])
            new_pixel = list(pixel)
            new_pixel[rgb_ind] = new_channel_val
            pxa[pixel_coords[0], pixel_coords[1]] = tuple(new_pixel)

    msg_len = len(byte_vals)
    encoded_msg_len_str = _msg_len_to_str(msg_len)

    return encoded_msg_len_str


def decode(image: Image, encoded_msg_len: str, start: int = DEFAULT_START, every_px: int = DEFAULT_EVERY_PX) -> str:
    byte_vals = []
    pxa = image.load()

    msg_len = _str_to_msg_len(encoded_msg_len)

    for i in range(msg_len):
        coords = _coords(i, start, every_px, image.width)
        rgb_ind = i % len(RGB)
        digits = [str(_decode_digit_from_channel_val(pxa[x, y][rgb_ind])) for x, y in coords]
        byte_val = int(''.join(digits))
        byte_vals.append(byte_val)

    bytes = bytearray(byte_vals)
    msg = bytes.decode()

    return msg



from typing import Optional

import typer

from stega import __app_name__, __version__
from stega.cli import textimg

app = typer.Typer()


def version_callback(value: bool):
    if value:
        typer.echo(f'{__app_name__} version: {__version__}')
        raise typer.Exit()


@app.callback(help='Stega-saurus image steganography')
def main_callback(
    version: Optional[bool] = typer.Option(None, '--version', callback=version_callback, is_eager=True)
):
    """
    Main callback method for steganography commands.
    """
    pass


app.add_typer(textimg.app, name='textimg', callback=main_callback)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="stega-saurus-py",
    version="0.0.1a5",
    author="Greg Allan",
    author_email="gregallandev@gmail.com",
    description="A steganography CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gregmallan/stega-saurus",
    project_urls={
        "Bug Tracker": "https://github.com/gregmallan/stega-saurus/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7.2",
    install_requires=[
        "Pillow>=9.1.1,<9.2",
        "typer>=0.4.1,<0.5",
    ],
    extras_require={
        "dev": [
            "pytest>=7.1.2,<7.2",
            "pytest-cov>=3.0.0,<3.1",
        ]
    }
)
